package org.example;

import org.apache.calcite.config.Lex;
import org.apache.calcite.jdbc.CalcitePrepare;
import org.apache.calcite.plan.RelOptUtil;
import org.apache.calcite.plan.hep.HepPlanner;
import org.apache.calcite.plan.hep.HepProgramBuilder;
import org.apache.calcite.rel.RelNode;
import org.apache.calcite.rel.RelRoot;
import org.apache.calcite.rel.rules.FilterJoinRule;
import org.apache.calcite.server.CalciteServerStatement;
import org.apache.calcite.sql.SqlExplainLevel;
import org.apache.calcite.sql.SqlNode;
import org.apache.calcite.sql.parser.SqlParser;
import org.apache.calcite.tools.FrameworkConfig;
import org.apache.calcite.tools.Frameworks;
import org.apache.calcite.tools.Planner;

import java.net.URL;
import java.net.URLDecoder;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;

public class OptimizerDemo1 {
    public static void main(String[] args) {
        String sql = "select * from TEST_CSV.TEST01 as t1 left join TEST_CSV.TEST02 as t2 " +
                "on t1.NAME1=t2.NAME3 " +
                "where t1.NAME1='hello'";
        String filePath = "/model.json";
        Connection connection = null;
        try {
            URL url = OptimizerDemo1.class.getResource(filePath);
            String str = URLDecoder.decode(url.toString(), "UTF-8");
            Properties info = new Properties();
            info.put("model", str.replace("file:", ""));
            connection = DriverManager.getConnection("jdbc:calcite:", info);
            RelRoot root = OptimizerDemo1.genRelRoot(connection, sql);
            OptimizerDemo1.optimize(root.rel);
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            try {
                connection.close();
            }catch (SQLException e){
                e.printStackTrace();
            }
        }
    }

    /**
     * 构建一个优化器，添加 FILTER_ON_JOIN 这个优化 rule，然后进行优化，打印优化前后对比，以及对应的 cost 信息
     * 在 Calcite 中，树上层的 Cost 信息都是通过下层的信息计算而来，所以最重要的就是底层 TableSacn 的 Cost 信息
     * 而默认的 rowcount 是 100，CPU 是 101
     * 可寻迹 Calcite 源码： TableScan.computeSelfCost(xxx) -> RelOptAbstractTable.getRowCount() 查看
     * @param rel
     * @return
     */
    public static RelNode optimize(RelNode rel){
        System.out.println("----------------- before optimizer ------------------");
        System.out.println(RelOptUtil.toString(rel, SqlExplainLevel.ALL_ATTRIBUTES));
        HepProgramBuilder builder = new HepProgramBuilder();
        builder.addRuleInstance(FilterJoinRule.FilterIntoJoinRule.Config.DEFAULT.toRule()); //note: 添加 rule
        HepPlanner hepPlanner = new HepPlanner(builder.build());  //同时也可以仿照 FilterIntoJoinRule 这个类实现自己的优化 rule
        hepPlanner.setRoot(rel);
        rel = hepPlanner.findBestExp();
        System.out.println("----------------- after optimizer ------------------");

        System.out.println(RelOptUtil.toString(rel, SqlExplainLevel.ALL_ATTRIBUTES));

        return rel;
    }

    /**
     * Planner解析，校验，然后生成RelNode，使用mysql的sql语法格式
     *
     * @param connection
     * @param sql
     * @return
     * @throws Exception 参考自：https://zhuanlan.zhihu.com/p/65345335
     */
    public static RelRoot genRelRoot(Connection connection, String sql) throws Exception {
        //从 conn 中获取相关的环境和配置，生成对应配置
        CalciteServerStatement st = connection.createStatement().unwrap(CalciteServerStatement.class);
        CalcitePrepare.Context prepareContext = st.createPrepareContext();
        final FrameworkConfig config = Frameworks.newConfigBuilder()
                .parserConfig(SqlParser.configBuilder().setLex(Lex.MYSQL).build())
                .defaultSchema(prepareContext.getRootSchema().plus())
                .build();
        Planner planner = Frameworks.getPlanner(config);
        RelRoot root = null;
        try {
            SqlNode parse1 = planner.parse(sql);
            SqlNode validate = planner.validate(parse1);
            root = planner.rel(validate);
        } catch (Exception e) {
            e.printStackTrace();
        }
        return root;
    }
}