package org.example;

import org.apache.calcite.avatica.util.Casing;
import org.apache.calcite.avatica.util.Quoting;
import org.apache.calcite.config.NullCollation;
import org.apache.calcite.sql.SqlDialect;
import org.apache.calcite.sql.SqlNode;
import org.apache.calcite.sql.SqlWriterConfig;
import org.apache.calcite.sql.dialect.*;
import org.apache.calcite.sql.parser.SqlParseException;
import org.apache.calcite.sql.parser.SqlParser;
import org.apache.calcite.sql.parser.impl.SqlParserImpl;
import org.apache.calcite.sql.pretty.SqlPrettyWriter;
import org.apache.calcite.sql.validate.SqlConformanceEnum;
import org.apache.calcite.util.SourceStringReader;
import org.apache.calcite.util.Util;

import java.io.Reader;
import java.util.Arrays;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.UnaryOperator;

public class CommonSqlParser {
    private final SqlWriterConfig SQL_WRITER_CONFIG =
            SqlPrettyWriter.config()
                    .withAlwaysUseParentheses(true)
                    .withUpdateSetListNewline(false)
                    .withFromFolding(SqlWriterConfig.LineFolding.TALL)
                    .withIndentation(0);
    private List<EngineType> tryParserList = Arrays.asList(EngineType.PRESTO,EngineType.HIVE,EngineType.SPARK);
    private UnaryOperator<SqlParser.Config> getTransform(
            SqlDialect dialect) {
        return dialect == null ? UnaryOperator.identity()
                : dialect::configureParser;
    }

    private SqlNode parseStmtAndHandleEx(String sql,
                                         UnaryOperator<SqlParser.Config> transform,
                                         Consumer<SqlParser> parserChecker,
                                         EngineType engineType) throws SqlParseException {
        final Reader reader = new SourceStringReader(sql);
        final SqlParser parser = getSqlParser(reader, transform, engineType);
        final SqlNode sqlNode;
        sqlNode = parser.parseStmt();
        parserChecker.accept(parser);
        return sqlNode;
    }

    private SqlParser getSqlParser(Reader source,
                                    UnaryOperator<SqlParser.Config> transform,
                                    EngineType engineType) {
        final SqlParser.Config config = getSqlParserConfig(transform, engineType);
        return SqlParser.create(source, config);
    }

    private SqlParser.Config getSqlParserConfig
            (UnaryOperator<SqlParser.Config> transform, EngineType engineType) {
        final SqlParser.Config configBuilder =
                SqlParser.config()
                        .withParserFactory(SqlParserImpl.FACTORY)
                        .withUnquotedCasing(Casing.UNCHANGED)
                        .withQuotedCasing(Casing.UNCHANGED)
                        .withCaseSensitive(false)
                        .withConformance(getDefaultSqlConformanceEnum(engineType));
        return transform.apply(configBuilder);
    }

    private SqlDialect getDefaultSqlDialect(EngineType engineType) {
        switch (engineType) {
            case HIVE:
            case TEZ:
                SqlDialect.Context hiveContext = SqlDialect.EMPTY_CONTEXT
                        .withDatabaseProduct(SqlDialect.DatabaseProduct.HIVE)
                        .withQuotedCasing(Casing.UNCHANGED)
                        .withUnquotedCasing(Casing.UNCHANGED)
                        .withNullCollation(NullCollation.LOW)
                        .withIdentifierQuoteString("`")
                        .withCaseSensitive(false);
                return new HiveSqlDialect(hiveContext);
//                return HiveSqlDialect.DEFAULT;
            case SPARK:
                return SparkSqlDialect.DEFAULT;
            case PRESTO:
                return PrestoSqlDialect.DEFAULT;
            default:
                return CalciteSqlDialect.DEFAULT;
        }
    }

    private SqlConformanceEnum getDefaultSqlConformanceEnum(EngineType engineType) {
        switch (engineType) {
            case HIVE:
            case TEZ:
                return SqlConformanceEnum.DEFAULT;
            case SPARK:
                return SqlConformanceEnum.DEFAULT;
            case PRESTO:
                return SqlConformanceEnum.PRESTO;
            default:
                return SqlConformanceEnum.DEFAULT;
        }
    }

    public SqlNode parserSql(String sql, EngineType userSpecified) throws SqlParseException {
        //todo if engine is auto    try for parser
        if (userSpecified.equals(EngineType.AUTO)) {
            for (EngineType engineType : tryParserList) {
                try {
                    final UnaryOperator<SqlParser.Config> transform = getTransform(
                            getDefaultSqlDialect(engineType));
                    return parseStmtAndHandleEx(sql, transform, parser -> {
                    }, userSpecified);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        } else {
            final UnaryOperator<SqlParser.Config> transform = getTransform(
                    getDefaultSqlDialect(userSpecified));
            return parseStmtAndHandleEx(sql, transform, parser -> {
            }, userSpecified);
        }
        throw new RuntimeException("parser sql error");
    }

    public String actualSql(SqlNode sqlNode, EngineType recommendEngine) {
        SqlDialect recommendSqlDialect = getDefaultSqlDialect(recommendEngine);
        final SqlDialect outDialect = Util.first(recommendSqlDialect, HiveSqlDialect.DEFAULT);
        final SqlWriterConfig writerConfig = SQL_WRITER_CONFIG.withDialect(outDialect);
        return sqlNode.toSqlString(c -> writerConfig).getSql();
    }

    public static void main(String[] args) throws SqlParseException {
        String sql = "select c1 from emps where `id` = 1";
        CommonSqlParser commonSqlParser = new CommonSqlParser();
        SqlNode sqlNode = commonSqlParser.parserSql(sql, EngineType.HIVE);
        String reWriteSql = commonSqlParser.actualSql(sqlNode, EngineType.PRESTO);
        System.out.println(reWriteSql);
    }
}
